print("import: 'zope'")
import zope

print("import: 'zope.interface'")
import zope.interface

print("import: 'zope.interface.common'")
import zope.interface.common

print("import: 'zope.interface.common.tests'")
import zope.interface.common.tests

print("import: 'zope.interface.common.mapping'")
import zope.interface.common.mapping

print("import: 'zope.interface._zope_interface_coptimizations'")
import zope.interface._zope_interface_coptimizations

print("import: 'zope.interface.tests'")
import zope.interface.tests

